/*
	Standard library file for PROGS SMS C library

	MEMUser handling library

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera, 14 October 1994
	23.09.1996 MEMUserZap(NULL) now possible
*/

#include "mem_h"
#include "err_h"
#include "job_h"

/* first some internal definitions of datatypes */

typedef struct list_ {
	struct list_ *next; 		/* next free segement */
	} list;

typedef struct {
	int ChunkSize;				/* number of bytes in a chunk */
	int BlockingFactor; 		/* number of chunks in a block */
	int BlockSize;				/* total size of a block */
	list *free; 				/* list of free chunks */
	list *blocks;				/* list of blocks with chunks */
	Job owner;					/* owner of user memory */
	} MEMUser;

/* Chunks should be defined as :
typedef MEMUser *Chunks;
	but I did not do this, because I want to hide the representation */

#define linkin(el, list) { (el)->next=list; list=el; }

#define linkout(el, list) { el=list; list=(el)->next; }


/* --- start of the actual routines --- */

Error MEMUserInit(Size BlockSize, int BlockingFactor, Chunks *id)
{
	MEMUser *mem;

	if (BlockingFactor<=0 || BlockSize<=0) return ERR_ORNG;
	if (!MEMAllocate(sizeof(MEMUser), &mem))
	{
		mem->ChunkSize=even(max(sizeof(list),BlockSize));
		mem->BlockingFactor=BlockingFactor;
		mem->BlockSize=mem->ChunkSize*mem->BlockingFactor+sizeof(list);
		JOBMe(&mem->owner);

		*id=(Chunks)mem;
		return ERR_OK;
	}
	else return ERR_IMEM;
}

Error MEMUserAllocate(Chunks id, void **base)
{
	MEMUser *mem=(MEMUser *)id;
	list *block, *chunk;
	int i;
	Error err=ERR_OK;

	if (!mem->free) 							/* if no free chunks left */
	{
		err=MEMAllocateFor(mem->BlockSize,
						   mem->owner, &block); /* allocate a new block */
		if (err) return err;					/* out of memory ?? */
		linkin(block, mem->blocks); 			/* extra block of Chunks */

		/* mark all chunks in the block as free */
		chunk=(list *)((char *)block+sizeof(list));
		for (i=mem->BlockingFactor; i; i--)
		{
			linkin(chunk,mem->free);
			chunk=(list *)((char *)chunk+mem->ChunkSize);
		}
	}

	/* return a free segment */
	linkout(chunk, mem->free);
	*base=(void *)chunk;
	return err;
}

Error MEMUserRelease(Chunks id, void *base)
{
	MEMUser *mem=(MEMUser *)id;

	if (base) linkin((list *)base, mem->free);
	return ERR_OK;
}

Error MEMUserZap(Chunks id)
{
	MEMUser *mem=(MEMUser *)id;
	list *block;

	if (mem==NULL) return ERR_OK;

	/* first we have to release all the blocks */
	while (block=mem->blocks)
	{
		mem->blocks=block->next;
		MEMRelease(block);
	}

	/* now release the MEMUser area (also handles ZapRoutine and ZapAddress) */
	return MEMRelease(mem);
}

Error MEMUserWhenZapped(Chunks id, void (*ZapRoutine)(void *base), char *ZapAddress)
{
	return MEMWhenZapped((void *)id, ZapRoutine, ZapAddress);
}
