/*
	Standard library file for PROGS SMS C library

	dir handling library
	DIROpen
	DIRNext
	DIRClose
	DIRSubOpen
	DIRSubNext
	DIRSubClose
	standard sieves

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera, started 6 August 1995
	24.05.1996 file length returned is now correct
	17.07.1997 DIROpen now appends "_" to dirname if not there
	26.12.1996 DIRSieveExtension now forces extension to be preceded by '_' or
			   '.' if it doesn't already start with either. Both are now
			   always allowed (but either is required).
*/

#include "dir.h"
#include "io.h"
#include "str.h"
#include "mem.h"
#include "ctype.h"

/* internal routines, data structures */

#define OPEN_DIR		(OpenKey)4

typedef struct {
	Size length;					/* file length */
	char access;					/* file access, currently always 0 */
	FileType type;					/* file type */
	Size dataspace; 				/* dataspace area */
	int unused; 					/* don't ask me */
	short namelen;					/* length of filename */
	char name[36];					/* filename, not '\0' terminated */
	Time Update;					/* update time */
	unsigned short version; 		/* version */
	short dontknow; 				/* don't know ?? */
	Time Backup;					/* backup time */
	} IOHeader;

typedef struct {
	Error (*sieve)(FileInfo *file, int parameter);
	int parameter;
	} Sieve;

typedef struct _toscan {
	struct _toscan *next;
	char dir[IO_MAXDIRECTORY];
	} ToScan;

#define TOSCANFACTOR 10 			/* blocking factor for ToScan Chunks */

typedef struct {
	Channel channel;				/* Channel opened with directory */
	char directory[IO_MAXDIRECTORY];/* base directory which is scanned */
	char device[IO_MAXDEVICENAME+4];/* device which contains directory */
	char tree, sub; 				/* scan tree ??, used for DIRSub ?? */
	ToScan *toscan; 				/* remaining subdirs to scan */
	Chunks toscanblocks;			/* chunk for subdirs to scan */
	int sievecount; 				/* number of sieves in this Iterator */
	Sieve sieves[1];				/* the array with the sieves */
	} Iterator;

/* internal routines, prototypes */
Error _IOHeaderGet(Channel chid, IOHeader *base);
static Error Open(char *directory, char tree, char sub, DirIterator *iterator,
				  Sieve *sieves);

/* the actual routines */

Error DIROpen(char *directory, char tree, DirIterator *iterator, ...)
{
	return Open(directory,tree,FALSE,iterator,(Sieve *)((&iterator)+1));
}

Error DIRSubOpen(char *directory, char tree, DirIterator *iterator)
{
	Sieve none={NULL,0};
	return Open(directory,tree,TRUE,iterator,&none);
}

Error DIRDirectory(DirIterator iterator, char *dirname)
{
	Iterator *it=(Iterator *)iterator;
	STRCopy(it->device,dirname);
	return STRAppendSafe(it->directory,dirname,IO_MAXDIRECTORY);
}

Error DIRNext(DirIterator iterator, FileInfo *result)
{
	Iterator *it=(Iterator *)iterator;
	IOHeader header;
	Error sieveres=ERR_IPAR;
	Error err;
	int i;
	Sieve *sieve;

	while (sieveres!=ERR_OK)
	{
		/* get next entry in directory */
		err=IOGetBytes(it->channel, sizeof(IOHeader), (char *)&header, NULL);
		if (err)
		{
			if (err==ERR_EOF && it->toscan)
			{
				ToScan *ts=it->toscan;
				/* restart scan with next ToScan object */
				IOClose(it->channel);
				err=IOOpen(ts->dir,OPEN_DIR,&it->channel);
				if (err) return err;
				it->toscan=ts->next;
				MEMUserRelease(it->toscanblocks,ts);
				continue;
			}
			else
				return err;
		}

		/* only use real entries in directory - not free space */
		if (header.length==0 && header.namelen==0)
			continue;

		/* fill in result */
		result->length=    header.length-64;	/* minus header length */
		result->type=	   header.type;
		result->dataspace= header.dataspace;
		result->UpdateTime=header.Update;
		result->BackupTime=header.Backup;
		result->version   =header.version;
		header.name[header.namelen]='\0';
		STRCopySafe(header.name+strlen(it->directory),result->name,IO_MAXNAME);

		if (result->type==FILETYPE_DIRECTORY)
		{
			if (it->tree)
			{
				/* store in ToScan object */
				ToScan *ts;
				err=MEMUserAllocate(it->toscanblocks,&ts);
				if (err) return err;
				STRCopy(it->device,ts->dir);
				STRAppendSafe(header.name,ts->dir,IO_MAXDIRECTORY);
				ts->next=it->toscan;
				it->toscan=ts;
			}

			if (!it->sub)
				continue;
		}
		else if (it->sub)
			continue;

		/* pass through sieve */
		sieveres=ERR_OK;			/* file is excepted up until now */
		for (i=it->sievecount, sieve=it->sieves ;
			 i && sieveres==ERR_OK ;
			 i--, sieve++)
		{
			sieveres=sieve->sieve(result,sieve->parameter);
		}
	}
	return err;
}

Error DIRClose(DirIterator iterator)
{
	Iterator *it=(Iterator *)iterator;

	IOClose(it->channel);
	MEMUserZap(it->toscanblocks);
	MEMRelease(it);

	return ERR_OK;
}


/* standard sieves */

Error DIRSieveLength(FileInfo *file, int parameter)
{
	if (file->length==parameter) return ERR_OK;
							else return ERR_IPAR;
}

Error DIRSieveLengthLarger(FileInfo *file, int parameter)
{
	if (file->length>=parameter) return ERR_OK;
							else return ERR_IPAR;
}

Error DIRSieveLengthSmaller(FileInfo *file, int parameter)
{
	if (file->length<=parameter) return ERR_OK;
							else return ERR_IPAR;
}

Error DIRSieveType(FileInfo *file, int parameter)
{
	if (file->type==parameter) return ERR_OK;
						  else return ERR_IPAR;
}

Error DIRSieveExtension(FileInfo *file, int parameter)
{
	int i, extlen;
	int namelen=strlen(file->name);
	char *start, *str, *ext=(char *)parameter;

	if (!*ext) return ERR_OK;		/* accept when no extensions to select */
	while (*ext)
	{
		/* find start and length of current extension to test */
		start=ext;
		for (extlen=0 ; *ext!=';' && *ext!='\0' ; extlen++,ext++);

		if (*ext==';') ext++;		/* skip separator */

		if (*start=='_' || *start=='.') start++, extlen--;
		if (namelen<extlen) continue;

		/* compare extension */
		str=file->name+namelen-extlen;
		if (str==file->name || str[-1]=='.' || str[-1]=='_')
			for( ; extlen && (tolower(*str++)==tolower(*start++)) ; extlen-- );
		if (!extlen) return ERR_OK;
	}
	return ERR_IPAR;
}

Error DIRSieveExtensionNot(FileInfo *file, int parameter)
{
	int i, extlen;
	int namelen=strlen(file->name);
	char *start, *str, *ext=(char *)parameter;

	if (!*ext) return ERR_OK;		/* accept when no extensions to select */
	while (*ext)
	{
		/* find start and length of current extension to test */
		start=ext;
		for (extlen=0 ; *ext!=';' && *ext!='\0' ; extlen++,ext++);

		if (*ext==';') ext++;		/* skip separator */

		if (namelen<extlen) continue;

		/* compare extension */
		for( str=file->name+namelen-extlen ;
			 extlen && (tolower(*str++)==tolower(*start++)) ;
			 extlen-- );
		if (!extlen) return ERR_IPAR;
	}
	return ERR_OK;
}


/* internal routines */

static Error Open(char *directory, char tree, char sub, DirIterator *iterator,
				  Sieve *sieves)
{
	Channel ch;
	Error err;
	Iterator *it;
	int count;
	IOHeader header;
	MediumInfo medium;
	char dir[IO_MAXDIRECTORY];

	/* make sure directory name ends in an underscore */
	STRCopySafe(directory,dir,IO_MAXDIRECTORY);
	if (dir[strlen(dir)-1]!='_') STRAppendSafe("_",dir,IO_MAXDIRECTORY);

	/* try to open directory */
	err=IOOpen(dir,OPEN_DIR,&ch);
	if (err) return err;

	/* figure out number of sieves */
	for (count=0 ; sieves[count].sieve ; count++);

	/* allocate Iterator and fill in */
	err=MEMAllocate(sizeof(Iterator)+sizeof(Sieve)*(count-1), &it);
	if (err) { IOClose(ch); return err; }
	MEMCopy(sizeof(Sieve)*count, sieves, it->sieves);
	it->sievecount=count;
	it->channel=ch;
	it->tree=tree;
	it->sub=sub;

	/* figure out name of this directory */
	err=_IOHeaderGet(ch, &header);
	if (err) { IOClose(ch); MEMRelease(it); return err; }
	header.name[header.namelen]='\0';
	STRCopySafe(header.name,it->directory,IO_MAXDIRECTORY);
	if (header.namelen) STRAppendSafe("_",it->directory,IO_MAXDIRECTORY);

	/* figure out medium name */
	if (isalpha(dir[0]) && isdigit(dir[1]) && dir[2]=='_')
		MEMCopy(3,directory,it->device);
	err=IOMedium(ch,&medium);
	if (err) { IOClose(ch); MEMRelease(it); return err; }
	STRAppend(medium.devicename,it->device);

	/* allocate Chunks for ToScan blocks */
	err=MEMUserInit(sizeof(ToScan),TOSCANFACTOR,&it->toscanblocks);
	if (err) { IOClose(ch); MEMRelease(it); return err; }

	/* fill in return value */
	*iterator=(DirIterator)it;
	return err;
}
